from __future__ import absolute_import

import warnings

from turbodbc_intern import connect as intern_connect
from turbodbc_intern import Rows

from .exceptions import translate_exceptions
from .connection import Connection

def _make_connection_string(dsn, **kwargs):
    if dsn:
        kwargs['dsn'] = dsn
    return ';'.join(["{}={}".format(key, value) for key, value in kwargs.iteritems()])


@translate_exceptions
def connect(dsn=None, read_buffer_size=None, rows_to_buffer=None, parameter_sets_to_buffer=None, use_async_io=False, **kwargs):
    """
    Create a connection with the database identified by the dsn
    :param dsn: data source name as given in the odbc.ini file
    :param rows_to_buffer: Number of rows which shall be buffered
                           when result sets are fetched from the server.
                           None means that the default value is used.
    :param parameter_sets_to_buffer: Number of parameter sets (rows) which
                           shall be buffered when bulk queries are sent
                           to the server.
                           None means that the default value is used.
    :param use_async_io: Set this value to true if you want to use
                         asynchronous I/O. Asynchronous I/O means that
                         while the main thread converts a batch of
                         results to Python objects, another thread
                         fetches additional results in the background.
    :param \**kwargs: You may specify additional options as you please.
                      These options will go into the connection string.
                      Valid options depend on the specific database you
                      would like to connect with (e.g. `user` or `password`)
    """    
    connection = Connection(intern_connect(_make_connection_string(dsn, **kwargs)))

    if rows_to_buffer:
        warnings.warn("Calling turbodbc.connect() with parameter rows_to_buffer is deprecated. "
                      "Instead, set the parameter read_buffer_size to turbodbc.Megabytes(x) or "
                      "turbodbc.Rows(y) instead.", DeprecationWarning)
        connection.impl.set_buffer_size(Rows(rows_to_buffer))

    if read_buffer_size:
        connection.impl.set_buffer_size(read_buffer_size)

    if parameter_sets_to_buffer:
        connection.impl.parameter_sets_to_buffer = parameter_sets_to_buffer

    connection.impl.use_async_io = use_async_io

    return connection


from __future__ import absolute_import

from .api_constants import apilevel, threadsafety, paramstyle
from .connect import connect
from .constructors import Date, Time, Timestamp
from .exceptions import Error, InterfaceError, DatabaseError
from .data_types import STRING, BINARY, NUMBER, DATETIME, ROWID
from turbodbc_intern import Rows, Megabytes





from setuptools import setup

import sys
import sysconfig
import os
import os.path as op
import distutils.spawn as ds
import distutils.dir_util as dd
import distutils.sysconfig
import distutils

import itertools
from glob import iglob


def _get_distutils_build_directory():
    """
    Returns the directory distutils uses to build its files.
    We need this directory since we build extensions which have to link
    other ones.
    """
    pattern = "lib.{platform}-{major}.{minor}"
    return os.path.join('build', pattern.format(platform=sysconfig.get_platform(),
                                                major=sys.version_info[0],
                                                minor=sys.version_info[1]))


def _get_source_files(directory):
    path = os.path.join('src', directory)
    iterable_sources = (iglob(os.path.join(root,'*.cpp')) for root, dirs, files in os.walk(path))
    source_files = itertools.chain.from_iterable(iterable_sources)
    return list(source_files)


def _remove_strict_prototype_option_from_distutils_config():
    strict_prototypes = '-Wstrict-prototypes'
    config = distutils.sysconfig.get_config_vars()
    for key, value in config.iteritems():
        if strict_prototypes in str(value):
            config[key] = config[key].replace(strict_prototypes, '')

_remove_strict_prototype_option_from_distutils_config()


def _has_numpy_headers():
    try:
        import numpy
        return True
    except:
        return False


def get_extension_modules():
    """
    Extension module which is actually a plain C++ library without Python bindings
    """
    turbodbc_sources = _get_source_files('cpp_odbc') + _get_source_files('turbodbc')
    turbodbc_library = distutils.core.Extension('libturbodbc',
                                                sources=turbodbc_sources,
                                                include_dirs=['include/'],
                                                extra_compile_args=['--std=c++11'],
                                                libraries=['odbc', 'boost_python'])
    
    """
    An extension module which contains the main Python bindings for turbodbc
    """
    turbodbc_python = distutils.core.Extension('turbodbc_intern',
                                               sources=_get_source_files('turbodbc_python'),
                                               include_dirs=['include/'],
                                               extra_compile_args=['--std=c++11'],
                                               libraries=['odbc', 'boost_python', 'turbodbc'],
                                               extra_link_args=["-Wl,-rpath,$ORIGIN"],
                                               library_dirs=[_get_distutils_build_directory()])
    
    """
    An extension module which contains Python bindings which require numpy support
    to work. Not included in the standard Python bindings so this can stay optional.
    """
    if _has_numpy_headers():
        import numpy
        turbodbc_numpy = distutils.core.Extension('turbodbc_numpy_support',
                                                  sources=_get_source_files('turbodbc_numpy'),
                                                  include_dirs=['include/', numpy.get_include()],
                                                  extra_compile_args=['--std=c++11'],
                                                  libraries=['odbc', 'boost_python', 'turbodbc'],
                                                  extra_link_args=["-Wl,-rpath,$ORIGIN"],
                                                  library_dirs=[_get_distutils_build_directory()])
        return [turbodbc_library, turbodbc_python, turbodbc_numpy]
    else:
        return [turbodbc_library, turbodbc_python]


setup(name = 'turbodbc',
      version = '0.4.1',
      description = 'turbodbc is a Python DB API 2.0 compatible ODBC driver',
      include_package_data = True,
      url = 'https://github.com/blue-yonder/turbodbc',
      author='Michael Koenig',
      author_email = 'michael.koenig@blue-yonder.com',
      packages = ['turbodbc'],
      classifiers = ['Development Status :: 4 - Beta',
                     'Intended Audience :: Developers',
                     'License :: OSI Approved :: MIT License',
                     'Operating System :: POSIX :: Linux',
                     'Programming Language :: C++',
                     'Programming Language :: Python :: 2.7',
                     'Topic :: Database'],
      ext_modules = get_extension_modules()
      )



from __future__ import absolute_import

from functools import wraps
from exceptions import StandardError

from turbodbc_intern import Error as InternError


class Error(StandardError):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass 


def translate_exceptions(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            return f(*args, **kwds)
        except InternError as e:
            raise DatabaseError(str(e))
    return wrapper


from __future__ import absolute_import

from itertools import islice
from collections import OrderedDict

from turbodbc_intern import has_result_set, make_row_based_result_set

from .exceptions import translate_exceptions, InterfaceError, Error


def _has_numpy_support():
    try:
        import turbodbc_numpy_support
        return True
    except ImportError:
        return False


def _make_masked_array(data, mask):
    from numpy.ma import MaskedArray
    from numpy import object_
    if isinstance(data, list):
        return MaskedArray(data=data, mask=mask, dtype=object_)
    else:
        return MaskedArray(data=data, mask=mask)


class Cursor(object):
    def __init__(self, impl):
        self.impl = impl
        self.result_set = None
        self.rowcount = -1
        self.arraysize = 1

    def __iter__(self):
        return self

    def next(self):
        element = self.fetchone()
        if element is None:
            raise StopIteration
        else:
            return element

    def _assert_valid(self):
        if self.impl is None:
            raise InterfaceError("Cursor already closed")

    def _assert_valid_result_set(self):
        if self.result_set is None:
            raise InterfaceError("No active result set")

    @property
    def description(self):
        if self.result_set:
            info = self.result_set.get_column_info()
            return [(c['name'], c['type_code'], None, None, None, None, c['supports_null_values']) for c in info]
        else:
            return None

    @translate_exceptions
    def execute(self, sql, parameters=None):
        """Execute an SQL query"""
        self._assert_valid()
        self.impl.prepare(sql)
        if parameters:
            # TODO: The list call here is probably inefficient
            # will go once the translation layer is better
            self.impl.add_parameter_set(list(parameters))
        self.impl.execute()
        self.rowcount = self.impl.get_row_count()
        cpp_result_set = self.impl.get_result_set()
        if has_result_set(cpp_result_set):
            self.result_set = make_row_based_result_set(cpp_result_set)
        else:
            self.result_set = None

    @translate_exceptions
    def executemany(self, sql, parameters=None):
        """Execute an SQL query"""
        self._assert_valid()
        self.impl.prepare(sql)

        if parameters:
            for parameter_set in parameters:
                # TODO: The list call here is probably inefficient
                # will go once the translation layer is better
                self.impl.add_parameter_set(list(parameter_set))

        self.impl.execute()
        self.rowcount = self.impl.get_row_count()
        cpp_result_set = self.impl.get_result_set()
        if has_result_set(cpp_result_set):
            self.result_set = make_row_based_result_set(cpp_result_set)
        else:
            self.result_set = None

    @translate_exceptions
    def fetchone(self):
        self._assert_valid_result_set()
        result = self.result_set.fetch_row()
        if len(result) == 0:
            return None 
        else:
            return result  

    @translate_exceptions    
    def fetchall(self):
        return [row for row in self]

    @translate_exceptions    
    def fetchmany(self, size=None):
        if size is None:
            size = self.arraysize
        if (size <= 0):
            raise InterfaceError("Invalid arraysize {} for fetchmany()".format(size))

        return [row for row in islice(self, size)]

    def fetchallnumpy(self):
        self._assert_valid_result_set()
        if _has_numpy_support():
            from turbodbc_numpy_support import make_numpy_result_set
            numpy_result_set = make_numpy_result_set(self.impl.get_result_set())
            column_names = [description[0] for description in self.description]
            columns = zip(column_names,
                          [_make_masked_array(data, mask) for data, mask in numpy_result_set.fetch_all()])
            return OrderedDict(columns)
        else:
            raise Error("turbodbc was compiled without numpy support. Please install "
                        "numpy and reinstall turbodbc")

    def close(self):
        self.result_set = None
        self.impl = None

    def setinputsizes(self, sizes):
        """
        setinputsizes() has no effect. turbodbc automatically picks appropriate
        return types and sizes. Method exists since PEP-249 requires it.
        """
        pass

    def setoutputsize(self, size, column=None):
        """
        setoutputsize() has no effect. turbodbc automatically picks appropriate
        input types and sizes. Method exists since PEP-249 requires it.
        """
        pass
